# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_about.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(439, 498)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.widget = QtWidgets.QWidget(Form)
        self.widget.setObjectName("widget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.widget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.banner = QtWidgets.QLabel(self.widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.banner.sizePolicy().hasHeightForWidth())
        self.banner.setSizePolicy(sizePolicy)
        self.banner.setMaximumSize(QtCore.QSize(130, 130))
        self.banner.setText("")
        self.banner.setPixmap(QtGui.QPixmap(":/hlp.png"))
        self.banner.setScaledContents(True)
        self.banner.setAlignment(QtCore.Qt.AlignCenter)
        self.banner.setObjectName("banner")
        self.horizontalLayout.addWidget(self.banner)
        self.verticalLayout.addWidget(self.widget)
        self.textBrowser = QtWidgets.QTextBrowser(Form)
        self.textBrowser.setOpenExternalLinks(True)
        self.textBrowser.setObjectName("textBrowser")
        self.verticalLayout.addWidget(self.textBrowser)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.textBrowser.setHtml(_translate("Form", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans\'; font-size:10pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">Idea and prototype by Hellzed.</span></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">Rewrite in Python by ngld.</span></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">For feedback and updates go to:<br /></span><a href=\"http://www.hard-light.net/forums/index.php?topic=86364\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt; text-decoration: underline; color:#0000ff;\">http://www.hard-light.net/forums/index.php?topic=86364</span></a></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">The code is available at:<br /></span><a href=\"https://github.com/ngld/Knossos\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt; text-decoration: underline; color:#0000ff;\">https://github.com/ngld/Knossos</span></a><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\"><br /></span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'Sans Serif\'; font-size:9pt;\"><br /></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">Dependencies:</span></p>\n"
"<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://python.org\"><span style=\" text-decoration: underline; color:#0000ff;\">Python</span></a> (2 or 3)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://qt-project.org/wiki/Category:LanguageBindings::PySide\"><span style=\" text-decoration: underline; color:#0000ff;\">PySide</span></a> or <a href=\"http://riverbankcomputing.co.uk/software/pyqt/intro\"><span style=\" text-decoration: underline; color:#0000ff;\">PyQt4</span></a> (UI)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://www.7-zip.org/\"><span style=\" text-decoration: underline; color:#0000ff;\">7-Zip</span></a> (extracts the downloaded archives)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"https://github.com/workhorsy/py-cpuinfo\"><span style=\" text-decoration: underline; color:#0000ff;\">py-cpuinfo</span></a> (detects CPU flags)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"https://pypi.python.org/pypi/semantic_version\"><span style=\" text-decoration: underline; color:#0000ff;\">semantic_version</span></a> (compares version numbers)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"https://pypi.python.org/pypi/six\"><span style=\" text-decoration: underline; color:#0000ff;\">six</span></a> (helps supporting Python 2 and 3 with a single codebase)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"https://pypi.python.org/pypi/requests\"><span style=\" text-decoration: underline; color:#0000ff;\">requests</span></a> (simplifies HTTP request handling)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://libsdl.org\"><span style=\" text-decoration: underline; color:#0000ff;\">SDL</span></a> (detects your graphic cards and joysticks)</li>\n"
"<li style=\" font-family:\'Sans Serif\'; font-size:9pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><a href=\"http://kcat.strangesoft.net/openal.html\"><span style=\" text-decoration: underline; color:#0000ff;\">OpenAL Soft</span></a> (detects your sound cards)</li></ul>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\"><br /></span></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">This tool also uses </span><a href=\"http://constexpr.org/innoextract/\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt; text-decoration: underline; color:#0000ff;\">InnoExtract</span></a><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\"> to unpack the GOG installer.</span></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Sans Serif\'; font-size:9pt;\">If you used the Windows installer or the DMG file for Mac OS, this program contains 7-Zip which is licensed under the GNU LGPL license.</span></p></body></html>"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_audio.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.playbackDevice = QtWidgets.QComboBox(Form)
        self.playbackDevice.setObjectName("playbackDevice")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.playbackDevice)
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.captureDevice = QtWidgets.QComboBox(Form)
        self.captureDevice.setObjectName("captureDevice")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.captureDevice)
        self.enableEFX = QtWidgets.QCheckBox(Form)
        self.enableEFX.setObjectName("enableEFX")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.enableEFX)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.sampleRate = QtWidgets.QSpinBox(Form)
        self.sampleRate.setMaximum(1000000)
        self.sampleRate.setSingleStep(100)
        self.sampleRate.setObjectName("sampleRate")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.sampleRate)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_2.setText(_translate("Form", "Playback device: "))
        self.label.setText(_translate("Form", "Capture device: "))
        self.enableEFX.setText(_translate("Form", "Enable EFX"))
        self.label_3.setText(_translate("Form", "Sample rate:"))
        self.sampleRate.setSuffix(_translate("Form", " Hz"))




#!/usr/bin/python
## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function
import sys

if __package__ is None and not hasattr(sys, 'frozen'):
    import os.path
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))


if len(sys.argv) > 1 and sys.argv[1] == '--cpuinfo':
    # We don't need to initialize knossos if we only need to fetch the CPU info.

    import json
    from knossos.third_party import cpuinfo

    info = None

    try:
        # Try querying the CPU cpuid register
        if not info and '--safe' not in sys.argv:
            info = cpuinfo.get_cpu_info_from_cpuid()

        # Try the Windows registry
        if not info:
            info = cpuinfo.get_cpu_info_from_registry()

        # Try /proc/cpuinfo
        if not info:
            info = cpuinfo.get_cpu_info_from_proc_cpuinfo()

        # Try sysctl
        if not info:
            info = cpuinfo.get_cpu_info_from_sysctl()

        # Try solaris
        if not info:
            info = cpuinfo.get_cpu_info_from_solaris()

        # Try dmesg
        if not info:
            info = cpuinfo.get_cpu_info_from_dmesg()
    except:
        from knossos.launcher import logging
        logging.exception('Failed to retrieve CPU info.')

    print(json.dumps(info))
else:
    from knossos import launcher
    launcher.main()



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import sys
from .qt import QtCore

# The version should follow the http://semver.org guidelines.
# Only remove the -dev tag if you're making a release!
VERSION = '0.1.2'
UPDATE_LINK = 'https://dev.tproxy.de/knossos'
INNOEXTRACT_LINK = 'https://dev.tproxy.de/knossos/innoextract.txt'
DEBUG = os.environ.get('KN_DEBUG') == '1'
SENTRY_DSN = 'https://77179552b41946488346a9a2d2669d74:f7b896367bd94f0ea960b8f0ee8b7a88@sentry.gruenprint.de/9'

app = None
main_win = None
shared_files = {}
fs2_watcher = None
pmaster = None
mods = None
installed = None
fso_flags = None
has_retail = None
raven = None

settings = {
    'fs2_bin': None,
    'fs2_path': None,
    'pins': {},
    'cmdlines': {},
    'hash_cache': None,
    'max_downloads': 3,
    'repos': [('https://fsnebula.org/repo/master.json', 'FSNebula')],
    'nebula_link': 'https://fsnebula.org/',
    'update_channel': 'stable',
    'update_notify': True,
    'ui_mode': 'hell',
    'keyboard_layout': 'default',
    'keyboard_setxkbmap': False,
    'use_raven': False
}

if '-dev' in VERSION:
    settings['update_channel'] = 'develop'

settings_path = os.path.expanduser('~/.knossos')
if sys.platform.startswith('win'):
    settings_path = os.path.expandvars('$APPDATA/knossos')


class _SignalContainer(QtCore.QObject):
    fs2_launched = QtCore.Signal()
    fs2_failed = QtCore.Signal(int)
    fs2_quit = QtCore.Signal()
    fs2_path_changed = QtCore.Signal()
    fs2_bin_changed = QtCore.Signal()
    list_updated = QtCore.Signal()
    repo_updated = QtCore.Signal()
    update_avail = QtCore.Signal('QVariant')
    task_launched = QtCore.Signal(QtCore.QObject)

signals = _SignalContainer()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/hell.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(499, 615)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.toolbar = QtWidgets.QFrame(self.centralwidget)
        self.toolbar.setObjectName("toolbar")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.toolbar)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pageControls = QtWidgets.QWidget(self.toolbar)
        self.pageControls.setObjectName("pageControls")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.pageControls)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.backButton = QtWidgets.QPushButton(self.pageControls)
        self.backButton.setObjectName("backButton")
        self.horizontalLayout_2.addWidget(self.backButton)
        self.horizontalLayout.addWidget(self.pageControls)
        self.listControls = QtWidgets.QWidget(self.toolbar)
        self.listControls.setObjectName("listControls")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.listControls)
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.searchEdit = QtWidgets.QLineEdit(self.listControls)
        self.searchEdit.setObjectName("searchEdit")
        self.horizontalLayout_3.addWidget(self.searchEdit)
        self.horizontalLayout.addWidget(self.listControls)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.updateButton = QtWidgets.QPushButton(self.toolbar)
        self.updateButton.setObjectName("updateButton")
        self.horizontalLayout.addWidget(self.updateButton)
        self.settingsButton = QtWidgets.QPushButton(self.toolbar)
        self.settingsButton.setObjectName("settingsButton")
        self.horizontalLayout.addWidget(self.settingsButton)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.toolbar)
        self.tabButtons = QtWidgets.QWidget(self.centralwidget)
        self.tabButtons.setObjectName("tabButtons")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.tabButtons)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.installedButton = QtWidgets.QToolButton(self.tabButtons)
        self.installedButton.setCheckable(True)
        self.installedButton.setAutoRaise(True)
        self.installedButton.setObjectName("installedButton")
        self.horizontalLayout_4.addWidget(self.installedButton)
        self.availableButton = QtWidgets.QToolButton(self.tabButtons)
        self.availableButton.setCheckable(True)
        self.availableButton.setAutoRaise(True)
        self.availableButton.setObjectName("availableButton")
        self.horizontalLayout_4.addWidget(self.availableButton)
        self.updatesButton = QtWidgets.QToolButton(self.tabButtons)
        self.updatesButton.setCheckable(True)
        self.updatesButton.setAutoRaise(True)
        self.updatesButton.setObjectName("updatesButton")
        self.horizontalLayout_4.addWidget(self.updatesButton)
        self.progressButton = QtWidgets.QToolButton(self.tabButtons)
        self.progressButton.setCheckable(True)
        self.progressButton.setAutoRaise(True)
        self.progressButton.setObjectName("progressButton")
        self.horizontalLayout_4.addWidget(self.progressButton)
        self.verticalLayout.addWidget(self.tabButtons)
        self.webView = QtWebKitWidgets.QWebView(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.webView.sizePolicy().hasHeightForWidth())
        self.webView.setSizePolicy(sizePolicy)
        self.webView.setUrl(QtCore.QUrl("about:blank"))
        self.webView.setObjectName("webView")
        self.verticalLayout.addWidget(self.webView)
        self.progressInfo = QtWidgets.QWidget(self.centralwidget)
        self.progressInfo.setObjectName("progressInfo")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.progressInfo)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.progressLabel = QtWidgets.QLabel(self.progressInfo)
        self.progressLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.progressLabel.setObjectName("progressLabel")
        self.verticalLayout_2.addWidget(self.progressLabel)
        self.progressBar = QtWidgets.QProgressBar(self.progressInfo)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_2.addWidget(self.progressBar)
        self.verticalLayout.addWidget(self.progressInfo)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 499, 23))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Knossos"))
        self.backButton.setText(_translate("MainWindow", "Back"))
        self.searchEdit.setPlaceholderText(_translate("MainWindow", "Search"))
        self.updateButton.setText(_translate("MainWindow", "Update List"))
        self.settingsButton.setText(_translate("MainWindow", "Settings"))
        self.installedButton.setText(_translate("MainWindow", "Installed"))
        self.availableButton.setText(_translate("MainWindow", "Available"))
        self.updatesButton.setText(_translate("MainWindow", "Updates"))
        self.progressButton.setText(_translate("MainWindow", "Progress"))
        self.progressLabel.setText(_translate("MainWindow", "TextLabel"))

from ..qt import QtWebKitWidgets



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_fso.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.build = QtWidgets.QComboBox(Form)
        self.build.setObjectName("build")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.build)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.fs2PathLabel = QtWidgets.QLabel(Form)
        self.fs2PathLabel.setObjectName("fs2PathLabel")
        self.horizontalLayout.addWidget(self.fs2PathLabel)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.browseButton = QtWidgets.QPushButton(Form)
        self.browseButton.setMaximumSize(QtCore.QSize(100, 16777215))
        self.browseButton.setObjectName("browseButton")
        self.horizontalLayout.addWidget(self.browseButton)
        self.formLayout.setLayout(1, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout)
        self.openLog = QtWidgets.QPushButton(Form)
        self.openLog.setObjectName("openLog")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.openLog)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "FS2 Path: "))
        self.label_2.setText(_translate("Form", "FSO Build: "))
        self.fs2PathLabel.setText(_translate("Form", "..."))
        self.browseButton.setText(_translate("Form", "Browse"))
        self.openLog.setText(_translate("Form", "Open fs2_open.log"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_knossos.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(398, 299)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.versionLabel = QtWidgets.QLabel(Form)
        self.versionLabel.setObjectName("versionLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.versionLabel)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.maxDownloads = QtWidgets.QSpinBox(Form)
        self.maxDownloads.setMaximum(5)
        self.maxDownloads.setObjectName("maxDownloads")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.maxDownloads)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.updateChannel = QtWidgets.QComboBox(Form)
        self.updateChannel.setObjectName("updateChannel")
        self.updateChannel.addItem("")
        self.updateChannel.addItem("")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.updateChannel)
        self.updateNotify = QtWidgets.QCheckBox(Form)
        self.updateNotify.setObjectName("updateNotify")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.updateNotify)
        self.debugLog = QtWidgets.QPushButton(Form)
        self.debugLog.setObjectName("debugLog")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.debugLog)
        self.clearHashes = QtWidgets.QPushButton(Form)
        self.clearHashes.setObjectName("clearHashes")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.clearHashes)
        self.label_5 = QtWidgets.QLabel(Form)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.reportErrors = QtWidgets.QCheckBox(Form)
        self.reportErrors.setObjectName("reportErrors")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.reportErrors)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Version:"))
        self.versionLabel.setText(_translate("Form", "?"))
        self.label_2.setText(_translate("Form", "Parallel downloads:"))
        self.label_3.setText(_translate("Form", "Update channel: "))
        self.updateChannel.setItemText(0, _translate("Form", "stable"))
        self.updateChannel.setItemText(1, _translate("Form", "develop"))
        self.updateNotify.setText(_translate("Form", "Display update notifications"))
        self.debugLog.setText(_translate("Form", "Open Knossos\' debug log"))
        self.clearHashes.setText(_translate("Form", "Clear the checksum cache"))
        self.label_5.setText(_translate("Form", "Troubleshooting:"))
        self.reportErrors.setText(_translate("Form", "Automatically report errors"))




## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import json
import shlex
import glob
import pickle

from . import uhf
uhf(__name__)

from . import center, util, repo, integration
from .qt import QtGui
from .tasks import run_task, CheckUpdateTask, CheckTask, FetchTask, InstallTask, UninstallTask
from .ui.select_list import Ui_Dialog as Ui_SelectList
from .windows import HellWindow, ModSettingsWindow, ModInstallWindow
from .repo import ModNotFound
from .ipc import IPCComm
from .runner import run_fs2, run_fs2_silent

# TODO: Split this file up into smaller parts and move them into the respective modules (i.e. run_mod should be in runner).

ipc_block = None


def save_settings():
    center.settings['hash_cache'] = dict()
    for path, info in util.HASH_CACHE.items():
        # Skip deleted files
        if os.path.exists(path):
            center.settings['hash_cache'][path] = info

    for mod in center.settings['cmdlines'].copy():
        if mod != '#default' and mod not in center.installed.mods:
            del center.settings['cmdlines'][mod]

    center.settings['pins'] = center.installed.pins

    with open(os.path.join(center.settings_path, 'settings.pick'), 'wb') as stream:
        pickle.dump(center.settings, stream, 2)


def select_fs2_path(interact=True):
    if interact:
        if center.settings['fs2_path'] is None:
            path = os.path.expanduser('~')
        else:
            path = center.settings['fs2_path']

        fs2_path = QtGui.QFileDialog.getExistingDirectory(center.main_win.win, 'Please select your FS2 directory.', path)
    else:
        fs2_path = center.settings['fs2_path']

    if fs2_path is not None and os.path.isdir(fs2_path):
        center.settings['fs2_path'] = os.path.abspath(fs2_path)

        bins = get_executables()
        if len(bins) == 1:
            # Found only one binary, select it by default.

            center.settings['fs2_bin'] = bins[0][1]
        elif len(bins) > 1:
            # Let the user choose.

            select_win = util.init_ui(Ui_SelectList(), QtGui.QDialog(center.main_win.win))
            has_default = False
            bins.sort()

            for i, path in enumerate(bins):
                select_win.listWidget.addItem(path[0])

                if not has_default and 'DEBUG' not in path[0]:
                    # Select the first non-debug build as default.

                    select_win.listWidget.setCurrentRow(i)
                    has_default = True

            select_win.listWidget.itemDoubleClicked.connect(select_win.accept)
            select_win.okButton.clicked.connect(select_win.accept)
            select_win.cancelButton.clicked.connect(select_win.reject)

            if select_win.exec_() == QtGui.QDialog.Accepted:
                center.settings['fs2_bin'] = bins[select_win.listWidget.currentRow()][1]

            select_win.deleteLater()
        else:
            center.settings['fs2_bin'] = None

        center.main_win.check_fso()
        center.signals.fs2_path_changed.emit()

        if center.settings['fs2_bin'] is not None:
            center.signals.fs2_bin_changed.emit()


def get_fso_flags():
    global fso_flags

    if center.settings['fs2_bin'] is None:
        return None

    if center.fso_flags is not None and center.fso_flags[0] == center.settings['fs2_bin']:
        return center.fso_flags[1]

    fs2_bin = os.path.join(center.settings['fs2_path'], center.settings['fs2_bin'])
    if not os.path.isfile(fs2_bin):
        return None

    if sys.platform.startswith('win'):
        flags_path = os.path.join(center.settings['fs2_path'], os.path.dirname(center.settings['fs2_bin']), 'flags.lch')
    else:
        flags_path = os.path.join(center.settings['fs2_path'], 'flags.lch')

    rc = run_fs2_silent(['-get_flags'])

    flags = None

    if rc != 1 and rc != 0:
        logging.error('Failed to run FSO! (Exit code was %d)', rc)
    elif not os.path.isfile(flags_path):
        logging.error('Could not find the flags file "%s"!', flags_path)
    else:
        with open(flags_path, 'rb') as stream:
            flags = util.FlagsReader(stream)

    center.fso_flags = (center.settings['fs2_bin'], flags)
    return flags


def get_executables():
    exes = []
    fs2_path = center.settings['fs2_path']

    for mid, mvs in center.installed.mods.items():
        for mod in mvs:
            for pkg in mod.packages:
                for item in pkg.executables:
                    name = mod.title + ' ' + item.get('version', '')
                    if item.get('debug', False):
                        name += ' (DEBUG)'

                    path = os.path.join(mod.folder, item['file'])
                    exes.append((name, path))

    if fs2_path is not None and os.path.isdir(fs2_path):
        fs2_path = os.path.abspath(fs2_path)
        if sys.platform == 'darwin':
            for app in glob.glob(os.path.join(fs2_path, '*.app')):
                name = os.path.basename(app)
                exes.append((name, os.path.join(name, 'Contents', 'MacOS', name[:-4])))
        else:
            bins = glob.glob(os.path.join(fs2_path, 'fs2_open_*'))

            for path in bins:
                path = os.path.basename(path)

                if not path.endswith(('.map', '.pdb')):
                    exes.append((path, path))

    return exes


def fetch_list():
    return run_task(FetchTask())


def get_cmdline(mod):
    if mod is None:
        return center.settings['cmdlines'].get('#default', [])[:]

    if mod.mid in center.settings['cmdlines']:
        return center.settings['cmdlines'][mod.mid][:]
    elif mod.cmdline != '':
        return shlex.split(mod.cmdline)
    else:
        return center.settings['cmdlines'].get('#default', [])[:]


def get_fso_profile_path():
    if sys.platform.startswith('linux'):
        return os.path.expanduser('~/.fs2_open')
    elif sys.platform == 'darwin':
        return os.path.expanduser('~/Library/FS2_Open')
    else:
        return center.settings['fs2_path']


def read_fso_cmdline():
    # Look for the cmdline path.
    path = get_fso_profile_path()
    if path is None:
        return []

    path = os.path.join(path, 'data/cmdline_fso.cfg')

    # Read the current cmdline.
    cmdline = []
    if os.path.exists(path):
        try:
            with open(path, 'r') as stream:
                cmdline = shlex.split(stream.read().strip())
        except:
            logging.exception('Failed to read "%s", assuming empty cmdline.', path)

    for i, part in enumerate(cmdline):
        if part.strip() == '-mod':
            del cmdline[i]

            if len(cmdline) > i:
                del cmdline[i]

            break

    return cmdline


def run_mod(mod):
    global installed

    if mod is None:
        mod = repo.Mod()

    modpath = util.ipath(os.path.join(center.settings['fs2_path'], mod.folder))
    mods = []

    def check_install():
        if not os.path.isdir(modpath) or mod.mid not in center.installed.mods:
            QtGui.QMessageBox.critical(center.app.activeWindow(), 'Error', 'Failed to install "%s"! Check the log for more information.' % (mod.title))
        else:
            run_mod(mod)

    if center.settings['fs2_bin'] is None:
        select_fs2_path()

        if center.settings['fs2_bin'] is None:
            QtGui.QMessageBox.critical(center.app.activeWindow(), 'Error', 'I couldn\'t find a FS2 executable. Can\'t run FS2!!')
            return

    try:
        inst_mod = center.installed.query(mod)
    except repo.ModNotFound:
        inst_mod = None

    if inst_mod is None:
        deps = center.mods.process_pkg_selection(mod.resolve_deps())
        titles = [pkg.name for pkg in deps if not center.installed.is_installed(pkg)]

        msg = QtGui.QMessageBox()
        msg.setIcon(QtGui.QMessageBox.Question)
        msg.setText('You don\'t have %s, yet. Shall I install it?' % (mod.title))
        msg.setInformativeText('%s will be installed.' % (', '.join(titles)))
        msg.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
        msg.setDefaultButton(QtGui.QMessageBox.Yes)

        if msg.exec_() == QtGui.QMessageBox.Yes:
            task = InstallTask(deps)
            task.done.connect(check_install)
            run_task(task)

        return

    try:
        mods = mod.get_mod_flag()
    except repo.ModNotFound as exc:
        QtGui.QMessageBox.critical(None, 'Knossos', 'Sorry, I can\'t start this mod because its dependency "%s" is missing!' % exc.mid)
        return

    if mods is None:
        return

    # Look for the cmdline path.
    path = os.path.join(get_fso_profile_path(), 'data/cmdline_fso.cfg')
    cmdline = get_cmdline(mod)

    if len(mods) == 0:
        for i, part in enumerate(cmdline):
            if part == '-mod':
                del cmdline[i]

                if len(cmdline) > i:
                    del cmdline[i]

                break

    elif '-mod' not in cmdline:
        cmdline.append('-mod')
        cmdline.append(','.join(mods))

    if not os.path.isfile(path):
        basep = os.path.dirname(path)
        if not os.path.isdir(basep):
            os.makedirs(basep)

    try:
        with open(path, 'w') as stream:
            stream.write(' '.join([shlex.quote(p) for p in cmdline]))
    except:
        logging.exception('Failed to modify "%s". Not starting FS2!!', path)

        QtGui.QMessageBox.critical(center.app.activeWindow(), 'Error', 'Failed to edit "%s"! I can\'t change the current mod!' % path)
    else:
        logging.info('Starting mod "%s" with cmdline "%s".', mod.title, cmdline)
        run_fs2()


def check_retail_files():
    if center.settings['fs2_path'] is None:
        return

    has_retail = False
    for item in os.listdir(center.settings['fs2_path']):
        if item.lower() == 'root_fs2.vp':
            has_retail = True
            break

    if has_retail:
        logging.debug('The FS2 path (%s) contains retail files!', center.settings['fs2_path'])
    else:
        logging.debug('The FS2 path (%s) does not contain retail files!', center.settings['fs2_path'])

    if has_retail != center.has_retail:
        center.has_retail = has_retail
        run_task(CheckTask())


##############
# Public API #
##############


def is_fso_installed():
    fs2_path = center.settings['fs2_path']
    if fs2_path is not None:
        fs2_bin = os.path.join(fs2_path, center.settings['fs2_bin'])

    return fs2_path is not None and fs2_bin is not None and os.path.isdir(fs2_path) and os.path.isfile(fs2_bin)


def get_mod(mid, version=None):
    if center.mods is None:
        QtGui.QMessageBox.critical(None, 'Knossos', 'Hmm... I never got a mod list. Get a coder!')
        return None
    else:
        try:
            return center.mods.query(mid, version)
        except ModNotFound:
            QtGui.QMessageBox.critical(None, 'Knossos', 'Mod "%s" could not be found!' % mid)
            return None


# TODO: Create a proper window for this and move it to the windows module.
def uninstall_pkgs(pkgs, name=None, cb=None):
    titles = [pkg.name for pkg in pkgs if center.installed.is_installed(pkg)]

    if name is None:
        name = 'these packages'

    msg = QtGui.QMessageBox()
    msg.setIcon(QtGui.QMessageBox.Question)
    msg.setText('Do you really want to uninstall %s?' % name)
    msg.setInformativeText('%s will be removed.' % (', '.join(titles)))
    msg.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
    msg.setDefaultButton(QtGui.QMessageBox.Yes)

    if msg.exec_() == QtGui.QMessageBox.Yes:
        task = UninstallTask(pkgs)
        if cb is not None:
            task.done.connect(cb)

        run_task(task)
        return True
    else:
        return False


def switch_ui_mode(nmode):
    old_win = center.main_win
    if nmode == 'hell':
        center.main_win = HellWindow()
    else:
        logging.error('Unknown UI mode "%s"! (Maybe you tried to use a legacy UI...)' % nmode)

    center.main_win.open()
    old_win.close()


#########
# Tools #
#########


def install_scheme_handler(interactive=True):
    logging.info('Installing scheme handler...')

    try:
        if integration.current.install_scheme_handler():
            if interactive:
                QtGui.QMessageBox.information(None, 'Knossos', 'Done!')
            return
    except:
        logging.exception('Failed to install the scheme handler!')

    QtGui.QMessageBox.critical(None, 'Knossos', 'I probably failed to install the scheme handler.\nRun me as administrator and try again.')


def setup_ipc():
    global ipc_block

    ipc_block = IPCComm(center.settings_path)
    ipc_block.messageReceived.connect(handle_ipc)
    ipc_block.listen()


def shutdown_ipc():
    global ipc_block

    ipc_block.close()


def handle_ipc(msg):
    msg = msg.data().decode('utf8', 'ignore').strip()

    try:
        msg = json.loads(msg)
    except:
        logging.exception('Failed to parse IPC message %s.', msg)
        return

    if msg[0] == 'focus':
        center.main_win.win.activateWindow()
        center.main_win.win.raise_()
    elif msg[0] == 'run':
        mod = get_mod(msg[1])

        if mod is not None:
            run_mod(mod)
    elif msg[0] == 'install':
        mod = get_mod(msg[1])
        pkgs = []

        if not mod:
            # TODO: Maybe we should update the mod DB here?
            return

        if len(msg) > 2:
            for pname in msg[2:]:
                for pkg in mod.packages:
                    if pkg.name == pname:
                        pkgs.append(pkg)

        center.main_win.win.activateWindow()

        if mod.mid not in center.installed.mods:
            ModInstallWindow(mod, pkgs)
        else:
            QtGui.QMessageBox.information(None, 'Knossos', 'Mod "%s" is already installed!' % (mod.title))
    elif msg[0] == 'settings':
        center.main_win.win.activateWindow()

        if len(msg) == 1 or msg[1] == '':
            center.main_win.show_fso_settings()
        else:
            mod = get_mod(msg[1])

            if mod is None or mod.mid not in center.installed.mods:
                if mod is None:
                    name = msg[1]
                else:
                    name = mod.title
                QtGui.QMessageBox.information(None, 'Knossos', 'Mod "%s" is not yet installed!' % (name))
            else:
                ModSettingsWindow(mod)
    else:
        QtGui.QMessageBox.critical(None, 'Knossos', 'The action "%s" is unknown!' % (msg[0]))


def _read_default_cmdline():
    if '#default' not in center.settings['cmdlines']:
        center.settings['cmdlines']['#default'] = read_fso_cmdline()


def enable_raven():
    from raven import Client

    center.raven = Client(center.SENTRY_DSN, release=center.VERSION)


def init_self():
    setup_ipc()
    center.signals.fs2_path_changed.connect(_read_default_cmdline)
    center.signals.fs2_path_changed.connect(check_retail_files)

    center.main_win.check_fso(False)
    center.main_win.update_mod_list()

    if center.settings['update_notify'] and not center.VERSION.endswith('-dev'):
        run_task(CheckUpdateTask())



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import logging
import ctypes.util


ENCODING = 'utf8'


class SDL_Rect(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_int16),
        ('y', ctypes.c_int16),
        ('w', ctypes.c_uint16),
        ('h', ctypes.c_uint16)
    ]


class SDL_DisplayMode(ctypes.Structure):
    _fields_ = [
        ('format', ctypes.c_int32),
        ('w', ctypes.c_int),
        ('h', ctypes.c_int),
        ('refresh_rate', ctypes.c_int),
        ('driverdata', ctypes.c_void_p)
    ]


class c_any_pointer(object):

    @classmethod
    def from_param(cls, val):
        return val


def load_lib(*names):
    exc = None
    
    for name in names:
        if '.' not in name:
            libname = ctypes.util.find_library(name)
            if libname is not None:
                name = libname
        try:
            return ctypes.cdll.LoadLibrary(name)
        except OSError as e:
            if exc is None:
                exc = e
    
    raise Exception(names[0] + ' could not be found!')


def double_zero_string(val):
    global alc, libc
    
    off = 0
    data = []
    while val and val[off]:
        if val[off] == b'\x00':
            break
        
        slen = libc.strlen(val)
        data.append(val[off:off + slen].decode(ENCODING, 'replace'))
        off += slen
    
    return data

libc = load_lib('c')

if sys.platform.startswith('linux'):
    # Make sure Xlib is thread-safe *before* we load Qt or SDL.
    try:
        xlib = load_lib('X11')
        xlib.XInitThreads()
    except:
        logging.exception('Failed to call XInitThreads().')

# Load SDL
try:
    sdl = load_lib('libSDL2-2.0.so.0', 'SDL2', 'SDL2.dll', 'libSDL2.dylib')
    SDL2 = True
except:
    # Try SDL 1.2
    sdl = load_lib('libSDL-1.2.so.0', 'SDL', 'SDL.dll', 'libSDL.dylib')
    SDL2 = False
 
# SDL constants
if SDL2:
    SDL_INIT_VIDEO = 0x00000020
    SDL_INIT_JOYSTICK = 0x00000200
    SDL_WINDOW_FULLSCREEN = 0x00000001
    
    # SDL.h
    sdl.SDL_InitSubSystem.argtypes = [ctypes.c_uint32]
    sdl.SDL_InitSubSystem.restype = ctypes.c_int
    
    sdl.SDL_QuitSubSystem.argtypes = [ctypes.c_uint32]
    sdl.SDL_QuitSubSystem.restype = None
    
    # SDL_error.h
    sdl.SDL_GetError.argtypes = []
    sdl.SDL_GetError.restype = ctypes.c_char_p
    
    # SDL_video.h
    sdl.SDL_VideoInit.argtypes = [ctypes.c_char_p]
    sdl.SDL_VideoInit.restype = ctypes.c_int
    
    sdl.SDL_VideoQuit.argtypes = []
    sdl.SDL_VideoQuit.restype = None
    
    sdl.SDL_GetNumVideoDisplays.argtypes = []
    sdl.SDL_GetNumVideoDisplays.restype = ctypes.c_int
    
    sdl.SDL_GetNumDisplayModes.argtypes = [ctypes.c_int]
    sdl.SDL_GetNumDisplayModes.restype = ctypes.c_int
    
    sdl.SDL_GetDisplayMode.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.POINTER(SDL_DisplayMode)]
    sdl.SDL_GetDisplayMode.restype = ctypes.c_int
    
    sdl.SDL_GetCurrentDisplayMode.argtypes = [ctypes.c_int, ctypes.POINTER(SDL_DisplayMode)]
    sdl.SDL_GetCurrentDisplayMode.restype = ctypes.c_int
    
    # SDL_joystick.h
    sdl.SDL_NumJoysticks.argtypes = []
    sdl.SDL_NumJoysticks.restype = ctypes.c_int
    
    sdl.SDL_JoystickNameForIndex.argtypes = [ctypes.c_int]
    sdl.SDL_JoystickNameForIndex.restype = ctypes.c_char_p
else:
    SDL_INIT_VIDEO = 0x00000020
    SDL_INIT_JOYSTICK = 0x00000200
    SDL_HWSURFACE = 0x00000001
    SDL_FULLSCREEN = 0x80000000
    
    sdl.SDL_InitSubSystem.argtypes = [ctypes.c_uint32]
    sdl.SDL_InitSubSystem.restype = ctypes.c_int
    
    sdl.SDL_QuitSubSystem.argtypes = [ctypes.c_uint32]
    sdl.SDL_QuitSubSystem.restype = None
    
    sdl.SDL_GetError.argtypes = []
    sdl.SDL_GetError.restype = ctypes.c_char_p
    
    sdl.SDL_ListModes.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
    sdl.SDL_ListModes.restype = ctypes.POINTER(ctypes.POINTER(SDL_Rect))
    
    sdl.SDL_NumJoysticks.argtypes = []
    sdl.SDL_NumJoysticks.restype = ctypes.c_int
    
    sdl.SDL_JoystickName.argtypes = [ctypes.c_int]
    sdl.SDL_JoystickName.restype = ctypes.c_char_p

# OpenAL constants
ALC_DEFAULT_DEVICE_SPECIFIER = 0x1004
ALC_DEVICE_SPECIFIER = 0x1005
ALC_CAPTURE_DEVICE_SPECIFIER = 0x310
ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER = 0x311

# Load OpenAL
alc = load_lib('libopenal.so.1.15.1', 'openal', 'OpenAL')
alc.alcIsExtensionPresent.restype = ctypes.c_bool
alc.alcGetString.restype = ctypes.POINTER(ctypes.c_char)

gtk = None
gobject = None


def init_gtk():
    global gtk, gobject

    if gtk:
        return True
    
    # Load GTK2
    try:
        gtk = load_lib('libgtk-x11-2.0.so.0', 'gtk-x11-2.0')
        gobject = load_lib('libgobject-2.0.so.0', 'gobject-2.0')
    except:
        logging.exception('Failed to load GTK!')

        # Maybe GTK isn't used.
        gtk = None
        gobject = None
    else:
        gtk.gtk_init_check.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.POINTER(ctypes.c_char_p))]
        gtk.gtk_init_check.restype = ctypes.c_bool

        gtk.gtk_settings_get_default.argtypes = []
        gtk.gtk_settings_get_default.restype = ctypes.c_void_p

        gobject.g_object_get.argtypes = [ctypes.c_void_p, ctypes.c_char_p, c_any_pointer, ctypes.c_void_p]
        gobject.g_object_get.restype = None

        gobject.g_free.argtypes = [ctypes.c_void_p]
        gobject.g_free.restype = None

        gobject.g_object_unref.argtypes = [ctypes.c_void_p]
        gobject.g_object_unref.restype = None

        if not gtk.gtk_init_check(None, None):
            logging.error('Failed to initialize GTK!')
        else:
            return True

    return False


if SDL2:
    def get_modes():
        if sdl.SDL_InitSubSystem(SDL_INIT_VIDEO) < 0 or sdl.SDL_VideoInit(None) < 0:
            logging.error('Failed to init SDL\'s video subsystem!')
            logging.error(sdl.SDL_GetError())
            return []
        
        modes = []
        for i in range(sdl.SDL_GetNumVideoDisplays()):
            for a in range(sdl.SDL_GetNumDisplayModes(i)):
                m = SDL_DisplayMode()
                sdl.SDL_GetDisplayMode(i, a, ctypes.byref(m))
                
                if (m.w, m.h) not in modes:
                    modes.append((m.w, m.h))
        
        sdl.SDL_VideoQuit()
        sdl.SDL_QuitSubSystem(SDL_INIT_VIDEO)
        return modes
    
    def list_joysticks():
        if sdl.SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0:
            logging.error('Failed to init SDL\'s joystick subsystem!')
            logging.error(sdl.SDL_GetError())
            return []
        
        joys = []
        for i in range(sdl.SDL_NumJoysticks()):
            joys.append(sdl.SDL_JoystickNameForIndex(i).decode(ENCODING))
        
        sdl.SDL_QuitSubSystem(SDL_INIT_JOYSTICK)
        return joys
else:
    def get_modes():
        try:
            if sdl.SDL_InitSubSystem(SDL_INIT_VIDEO) < 0:
                logging.error('Failed to init SDL\'s video subsystem!')
                logging.error(sdl.SDL_GetError())
                return []
            
            modes = sdl.SDL_ListModes(None, SDL_FULLSCREEN | SDL_HWSURFACE)
            my_modes = []
            
            for mode in modes:
                if not mode:
                    break
                
                rect = mode[0]
                my_modes.append((rect.w, rect.h))
            
            sdl.SDL_QuitSubSystem(SDL_INIT_VIDEO)
            return my_modes
        except:
            logging.exception('Failed to call SDL_ListModes()!')
            return []

    def list_joysticks():
        try:
            if sdl.SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0:
                logging.error('Failed to init SDL\'s joystick subsystem!')
                logging.error(sdl.SDL_GetError())
                return []
            
            joys = []
            for i in range(sdl.SDL_NumJoysticks()):
                joys.append(sdl.SDL_JoystickName(i).decode(ENCODING, 'replace'))
            
            sdl.SDL_QuitSubSystem(SDL_INIT_JOYSTICK)
            return joys
        except:
            logging.exception('Failed to ask SDL for joysticks!')
            return []


def can_detect_audio():
    # OpenAL Soft doesn't support ALC_ENUMERATION_EXT but still provides a reasonable
    # list. For now, I'll pretend this extension is always present.
    return True  # alc.alcIsExtensionPresent(None, 'ALC_ENUMERATION_EXT')


def list_audio_devs():
    devs = double_zero_string(alc.alcGetString(None, ALC_DEVICE_SPECIFIER))
    default = alc.alcGetString(None, ALC_DEFAULT_DEVICE_SPECIFIER)
    
    captures = double_zero_string(alc.alcGetString(None, ALC_CAPTURE_DEVICE_SPECIFIER))
    default_capture = alc.alcGetString(None, ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER)
    
    default = ctypes.cast(default, ctypes.c_char_p).value.decode(ENCODING, 'replace')
    default_capture = ctypes.cast(default_capture, ctypes.c_char_p).value.decode(ENCODING, 'replace')
    
    return devs, default, captures, default_capture


def g_object_get_string(obj, prop):
    prop = ctypes.c_char_p(prop.encode('utf8'))
    value = ctypes.c_char_p()

    gobject.g_object_get(obj, prop, ctypes.byref(value), 0)

    py_value = value.value
    if py_value:
        py_value = py_value.decode('utf8', 'replace')

    gobject.g_free(value)
    return py_value


def get_gtk_theme():
    global gtk, gobject

    if not gtk:
        return ''

    settings = gtk.gtk_settings_get_default()
    if settings:
        return g_object_get_string(settings, 'gtk-theme-name')
    else:
        return None



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_versions.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Nothing to see here... move along."))




#!/usr/bin/python
## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from setuptools import setup
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='knossos',

    # This version should comply with PEP440 (http://legacy.python.org/dev/peps/pep-0440/).
    # The first three numbers should be the same as VERSION in knossos/center.py.
    version='0.1.2',

    description='A simple mod manager for FreeSpace 2 Open',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/ngld/knossos',

    # Author details
    author='ngld',
    author_email='ngld@tproxy.de',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: Apache Software License',

        # Supported Python versions
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],

    keywords='fso freespace',
    packages=['knossos', 'knossos.ui', 'knossos.third_party'],
    install_requires=['six', 'requests', 'semantic_version', 'raven', 'PyQt5'],

    # List additional groups of dependencies here (e.g. development dependencies).
    # You can install these using the following syntax, for example:
    # $ pip install -e .[dev,test]
    extras_require={},

    package_data={
        'knossos': ['data/*']
    },

    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'gui_scripts': [
            'knossos=knossos.launcher:main',
        ]
    }
)



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging

default_variant = 'auto'
QtCore = None

variant = os.environ.get('QT_API', default_variant)
if variant not in ('PyQt5', 'headless', default_variant):
    logging.warning('Unknown QT_API "%s"! Using default...', variant)
    variant = default_variant

if variant != 'headless':
    # Make sure we initialize Xlib before we load Qt.
    from . import clibs

if False and variant in ('PySide', 'auto'):
    try:
        from PySide import QtWebKit, QtCore, QtGui, QtNetwork

        # Success!
        variant = 'PySide'
    except ImportError:
        logging.exception('I was unable to load Qt! Tried PySide.')

        # If variant is 'auto', we fallback to PyQt4.
        if variant != 'auto':
            sys.exit(1)

if variant in ('PyQt5', 'auto'):
    try:
        #import sip
        #api2_classes = [
        #    'QData', 'QDateTime', 'QString', 'QTextStream',
        #    'QTime', 'QUrl', 'QVariant',
        #]

        #for cl in api2_classes:
        #    sip.setapi(cl, 2)

        from PyQt5 import QtCore, QtGui, QtWidgets, QtNetwork, QtWebKit, QtWebKitWidgets

        QtCore.Signal = QtCore.pyqtSignal
        QtCore.Slot = QtCore.pyqtSlot
        QtCore.QString = str

        # Success!
        variant = 'PyQt5'

    except ImportError:
        logging.exception('I was unable to load Qt! Tried PyQt4.')
        sys.exit(1)

if variant == 'headless':
    # This is just a dummy implementation of the QtCore.Signal() system. Nothing else is provided by this variant.
    import threading
    import time

    _main_loop = None

    class _App(object):
        _running = True
        _queue = None
        _queue_lock = None
        _timers = None

        def __init__(self, args):
            global _main_loop

            _main_loop = self
            self._queue = []
            self._queue_lock = threading.Lock()
            self._timers = []

        def reset(self):
            self._queue = []
            self._timers = []
            self._running = True

        def schedule(self, cb, args):
            with self._queue_lock:
                self._queue.append((cb, args))

        def set_timer(self, cb, timeout):
            with self._queue_lock:
                self._timers.append((cb, time.time() + timeout))

        def quit(self):
            self._running = False

        def exit(self, code=0):
            self._running = False

        def exec_(self):
            while self._running:
                time.sleep(0.3)
                ct = time.time()

                with self._queue_lock:
                    q = self._queue
                    self._queue = []

                    for i in range(len(self._timers) - 1, -1, -1):
                        if self._timers[i][1] < ct:
                            q.append((self._timers[i][0], []))
                            del self._timers[i]

                for cb, args in q:
                    cb(*args)

    class _Signal(object):
        _listeners = None
        _lock = None
        _active = False

        def __init__(self, *argtypes):
            # NOTE: Since this is just a dummy implementation, I won't check the argtypes...
            pass

        def _init(self):
            inst = _Signal()
            inst._listeners = []
            inst._lock = threading.Lock()
            return inst

        def connect(self, cb):
            if self._lock is None:
                raise Exception('QtCore.Signal() was not initialized correctly!')

            with self._lock:
                self._listeners.append(cb)

        def disconnect(self, cb):
            if self._lock is None:
                raise Exception('QtCore.Signal() was not initialized correctly!')

            with self._lock:
                self._listeners.remove(cb)

        def emit(self, *args):
            global _main_loop

            if self._lock is None:
                raise Exception('QtCore.Signal() was not initialized correctly!')

            with self._lock:
                for cb in self._listeners:
                    _main_loop.schedule(cb, args)

    class _QObject(object):

        def __init__(self):
            super(_QObject, self).__init__()

            # Init this class's signals for every instance.
            done = []
            bases = list(self.__class__.__bases__)
            bases.append(self.__class__)
            for base in bases:
                for name in base.__dict__:
                    if name not in done:
                        val = getattr(self, name)
                        done.append(name)

                        if isinstance(val, _Signal):
                            setattr(self, name, val._init())

    class _QTimer(object):

        @staticmethod
        def singleShot(timeout, cb):
            global _main_loop
            _main_loop.set_timer(cb, timeout / 1000.)

    class QtCore(object):
        QObject = _QObject
        QCoreApplication = _App
        Signal = _Signal
        QTimer = _QTimer
        QByteArray = None

    class QtGui(object):
        QApplication = _App
        QDialog = _QObject

    QtNetwork = None
    QtWebKit = None


def read_file(path):
    fd = QtCore.QFile(path)
    fd.open(QtCore.QIODevice.ReadOnly)
    data = str(fd.readAll())
    fd.close()

    return data


def load_styles(*names):
    data = ''
    for name in names:
        if ':/' in name:
            sheet = read_file(name)
        else:
            with open(name, 'r') as stream:
                sheet = stream.read()

        data += sheet.replace('url(./', 'url(' + os.path.dirname(name) + '/')

    return data

__all__ = ['QtCore', 'QtGui', 'QtNetwork', 'QtWebKit', 'variant', 'read_file', 'load_styles']



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import subprocess
import time
import json
import traceback
import six
from six.moves.urllib import parse as urlparse

logging.basicConfig(level=logging.DEBUG, format='%(levelname)s:%(threadName)s:%(module)s.%(funcName)s: %(message)s')

# We have to be in the correct directory *before* we import clibs so we're going to do this as early as possible.
if hasattr(sys, 'frozen'):
    if hasattr(sys, '_MEIPASS'):
        os.chdir(sys._MEIPASS)
    else:
        os.chdir(os.path.dirname(sys.executable))
else:
    my_path = os.path.dirname(__file__)
    if my_path != '':
        os.chdir(my_path)


from . import center

# Initialize the FileHandler early to capture all log messages.
if not os.path.isdir(center.settings_path):
    os.makedirs(center.settings_path)

log_path = None
if sys.platform.startswith('win'):
    # Windows won't display a console. Let's write our log messages to a file.
    # We truncate the log file on every start to avoid filling the user's disk with useless data.
    log_path = os.path.join(center.settings_path, 'log.txt')

    try:
        if os.path.isfile(log_path):
            os.unlink(log_path)
    except:
        # This will only be visible if the user is running a console version.
        logging.exception('The log is in use by someone!')
    else:
        handler = logging.FileHandler(log_path, 'w')
        handler.setFormatter(logging.Formatter('%(levelname)s:%(threadName)s:%(module)s.%(funcName)s: %(message)s'))
        handler.setLevel(logging.DEBUG)
        logging.getLogger().addHandler(handler)

if not center.DEBUG:
    logging.getLogger().setLevel(logging.INFO)

if six.PY2:
    from . import py2_compat

from .qt import QtCore, QtGui, QtWidgets, read_file, variant as qt_variant
from .ipc import IPCComm
from . import util


app = None
ipc = None


def handle_error():
    global app, ipc
    # TODO: Try again?

    QtWidgets.QMessageBox.critical(None, 'Knossos', 'Failed to connect to main process!')
    if ipc is not None:
        ipc.clean()

    app.quit()


def my_excepthook(type, value, tb):
    # NOTE: We can't use logging.exception() here because it uses traceback.print_exception
    # which (for some reason) doesn't work here.
    logging.error('UNCAUGHT EXCEPTION!\n%s%s: %s', ''.join(traceback.format_tb(tb)), type.__name__, value)
    if center.raven:
        center.raven.captureException((type, value, tb))


def get_cmd(args=[]):
    if hasattr(sys, 'frozen') and sys.frozen == 1:
        my_path = [os.path.abspath(sys.executable)]
    else:
        my_path = os.path.realpath(os.path.abspath(__file__))
        my_path = [os.path.abspath(sys.executable), os.path.join(os.path.dirname(my_path), '__main__.py')]

    return my_path + args


def get_file_path(name):
    if hasattr(sys, 'frozen') and sys.frozen == 1:
        return os.path.join('data', name)

    my_path = os.path.dirname(os.path.realpath(os.path.abspath(__file__)))
    data_path = os.path.join(my_path, 'data')
    if os.path.isdir(data_path):
        return os.path.join(data_path, name)
    else:
        from pkg_resources import resource_filename
        return resource_filename(__package__, name)


def run_knossos():
    global app

    import pickle

    from . import repo, progress, api
    from .windows import HellWindow

    if not util.test_7z():
        QtWidgets.QMessageBox.critical(None, 'Error', 'I can\'t find "7z"! Please install it and run this program again.', QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)
        return

    # Try to load our settings.
    spath = os.path.join(center.settings_path, 'settings.pick')
    settings = center.settings
    if os.path.exists(spath):
        defaults = settings.copy()

        try:
            with open(spath, 'rb') as stream:
                if six.PY3:
                    settings.update(pickle.load(stream, encoding='utf8', errors='replace'))
                else:
                    settings.update(pickle.load(stream))
        except:
            logging.exception('Failed to load settings from "%s"!', spath)

        # Migration
        if 's_version' not in settings or settings['s_version'] < 2:
            settings['repos'] = defaults['repos']
            settings['s_version'] = 2

        if settings['s_version'] < 3:
            if 'mods' in settings:
                del settings['mods']
            if 'installed_mods' in settings:
                del settings['installed_mods']

            settings['s_version'] = 3

        if settings['s_version'] < 4:
            settings['repos'] = defaults['repos']
            settings['s_version'] = 4

        del defaults
    else:
        # Most recent settings version
        settings['s_version'] = 4

    if '#default' not in settings['cmdlines']:
        settings['cmdlines']['#default'] = api.read_fso_cmdline()

    if settings['hash_cache'] is not None:
        util.HASH_CACHE = settings['hash_cache']

    if settings['fs2_path'] is None:
        settings['fs2_bin'] = None
    else:
        if not os.path.isdir(settings['fs2_path']):
            settings['fs2_bin'] = None
            settings['fs2_path'] = None
        elif settings['fs2_bin'] is not None:
            if not os.path.isfile(os.path.join(settings['fs2_path'], settings['fs2_bin'])):
                settings['fs2_bin'] = None

    if settings['use_raven']:
        api.enable_raven()

    util.DL_POOL.set_capacity(settings['max_downloads'])

    center.app = app

    center.installed = repo.InstalledRepo()
    center.installed.pins = settings['pins']
    center.pmaster = progress.Master()
    center.pmaster.start_workers(10)
    center.mods = repo.Repo()

    api.check_retail_files()
    mod_db = os.path.join(center.settings_path, 'mods.json')
    if os.path.isfile(mod_db):
        center.mods.load_json(mod_db)

    QtGui.QFontDatabase.addApplicationFont(':/html/fonts/FontAwesome.otf')

    if sys.platform == 'darwin':
        # QFontDatabase.addApplicationFont apparently doesn't work on Mac OS.
        # We work around this issue by copying the font to ~/Library/Fonts and deleting it once we exit.
        # FIXME: Solve the actual problem and remove this workaround.
        font_path = os.path.expandvars('$HOME/Library/Fonts/__tmp_Knossos_FontAwesome.otf')
        with open(font_path, 'wb') as stream:
            stream.write(read_file(':/html/fonts/FontAwesome.otf'))

        import atexit
        atexit.register(lambda: os.unlink(font_path))

    center.main_win = HellWindow()
    QtCore.QTimer.singleShot(1, api.init_self)

    center.main_win.open()
    app.exec_()

    api.save_settings()
    api.shutdown_ipc()


def scheme_handler(link):
    global app, ipc

    if not link.startswith(('fs2://', 'fso://')):
        # NOTE: fs2:// is deprecated, we don't tell anyone about it.
        QtWidgets.QMessageBox.critical(None, 'Knossos', 'I don\'t know how to handle "%s"! I only know fso:// .' % (link))
        app.quit()
        return

    link = urlparse.unquote(link.strip()).split('/')

    if len(link) < 3:
        QtWidgets.QMessageBox.critical(None, 'Knossos', 'Not enough arguments!')
        app.quit()
        return

    if not ipc.server_exists():
        # Launch the program.
        subprocess.Popen(get_cmd())

        # Wait for the program...
        start = time.time()
        while not ipc.server_exists():
            if time.time() - start > 20:
                # That's too long!
                QtWidgets.QMessageBox.critical(None, 'Knossos', 'Failed to start server!')
                app.quit()
                return

            time.sleep(0.3)

    try:
        ipc.open_connection(handle_error)
    except:
        logging.exception('Failed to connect to myself!')
        handle_error()
        return

    ipc.send_message(json.dumps(link[2:]))
    ipc.close(True)
    app.quit()
    return


def get_cpu_info():
    from .third_party import cpuinfo

    info = None

    try:
        # Try querying the CPU cpuid register
        if not info and '--safe' not in sys.argv:
            info = cpuinfo.get_cpu_info_from_cpuid()

        # Try the Windows registry
        if not info:
            info = cpuinfo.get_cpu_info_from_registry()

        # Try /proc/cpuinfo
        if not info:
            info = cpuinfo.get_cpu_info_from_proc_cpuinfo()

        # Try sysctl
        if not info:
            info = cpuinfo.get_cpu_info_from_sysctl()

        # Try solaris
        if not info:
            info = cpuinfo.get_cpu_info_from_solaris()

        # Try dmesg
        if not info:
            info = cpuinfo.get_cpu_info_from_dmesg()
    except:
        logging.exception('Failed to retrieve CPU info.')

    print(json.dumps(info))


def init():
    sys.excepthook = my_excepthook

    from . import integration

    if sys.platform.startswith('win') and os.path.isfile('7z.exe'):
        util.SEVEN_PATH = os.path.abspath('7z.exe')
    elif sys.platform == 'darwin' and os.path.isfile('7z'):
        util.SEVEN_PATH = os.path.abspath('7z')

    # The version file is only read in dev builds.
    if center.VERSION.endswith('-dev') and os.path.isfile('version'):
        with open('version', 'r') as data:
            version = data.read().strip()

            if version != center.VERSION:
                if version.startswith(center.VERSION):
                    center.VERSION = version
                else:
                    logging.error('Found invalid version file! The file contains %s but I\'m %s.', version, center.VERSION)

    logging.info('Running Knossos %s on %s.', center.VERSION, qt_variant)
    app = QtWidgets.QApplication([])

    logging.debug('Loading resources from %s.', get_file_path('resources.rcc'))
    QtCore.QResource.registerResource(get_file_path('resources.rcc'))

    app.setWindowIcon(QtGui.QIcon(':/hlp.png'))
    integration.init()

    return app


def main():
    global ipc, app

    # Default to replace errors when de/encoding.
    import codecs

    codecs.register_error('strict', codecs.replace_errors)
    codecs.register_error('really_strict', codecs.strict_errors)

    if len(sys.argv) > 1 and sys.argv[1] == '--cpuinfo':
        get_cpu_info()
        return

    app = init()
    ipc = IPCComm(center.settings_path)

    if len(sys.argv) > 1:
        if sys.argv[1] == '--install-scheme':
            from . import api

            api.install_scheme_handler('--silent' not in sys.argv)
            return
        elif sys.argv[1] == '--finish-update':
            updater = sys.argv[2]

            if not os.path.isfile(updater):
                logging.error('The update finished but where is the installer?! It\'s not where it\'s supposed to be! (%s)', updater)
            else:
                tries = 3
                while tries > 0:
                    try:
                        # Clean up
                        os.unlink(updater)
                    except:
                        logging.exception('Failed to remove updater! (%s)' % updater)

                    if os.path.isfile(updater):
                        time.sleep(0.3)
                        tries -= 1
                    else:
                        break

                # Delete the temporary directory.
                if os.path.basename(updater) == 'knossos_updater.exe':
                    try:
                        os.rmdir(os.path.dirname(updater))
                    except:
                        logging.exception('Failed to remove the updater\'s temporary directory.')
        else:
            scheme_handler(sys.argv[1])
            return
    elif ipc.server_exists():
        scheme_handler('fso://focus')
        return

    del ipc

    try:
        run_knossos()
    except:
        logging.exception('Uncaught exeception! Quitting...')

        # Try to tell the user
        QtWidgets.QMessageBox.critical(None, 'Knossos', 'I encountered a fatal error.\nI\'m sorry but I\'m going to crash now...')



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import logging
import threading
import six

from .qt import QtCore

try:
    import curses
except ImportError:
    curses = None

if six.PY2:
    threading.get_ident = lambda: threading.current_thread().ident

_progress = threading.local()


def reset():
    global _progress
    
    _progress.value = 0.0
    _progress.text = ''
    _progress.tasks = []
    _progress.callback = None

# Initialize with empty values.
reset()


# total_progress = off + span * task_progress
# Example: If you're tracking downloads and starting file 3 of 10, call start_task(3/10, 'Downloading whatever (%s)...')
def start_task(off, span, tmpl='%s'):
    if not hasattr(_progress, 'tasks'):
        reset()
    
    _progress.tasks.insert(0, (off, float(span), tmpl))


def finish_task():
    if not hasattr(_progress, 'tasks'):
        reset()
    
    _progress.tasks.pop(0)


def set_callback(cb):
    if not hasattr(_progress, 'tasks'):
        reset()
    
    _progress.callback = cb


def update(prog, text=''):
    global _progress
    
    if not hasattr(_progress, 'tasks'):
        reset()
    
    for task in _progress.tasks:
        prog = task[0] + prog * task[1]
        if '%s' in task[2]:
            text = task[2] % (text,)
        else:
            text = task[2]
    
    _progress.value = prog
    _progress.text = text
    
    if _progress.callback is not None:
        _progress.callback(prog, text)


# Task scheduler
class Worker(threading.Thread):

    def __init__(self, master):
        super(Worker, self).__init__()
        
        self._master = master
        self.daemon = True
        self.start()
    
    def run(self):
        while True:
            task = self._master._get_work()
            if task is None:
                return
            
            try:
                reset()
                set_callback(task[0]._track_progress)
                task[0]._init()
                task[0].work(*task[1])
            except SystemExit:
                return
            except:
                logging.exception('Exception in Thread!')
            
            task[0]._deinit()


class Master(object):
    _tasks = None
    _tasks_lock = None
    _workers = None
    _stop_workers = False
    _worker_cond = None
    
    def __init__(self):
        self._tasks = []
        self._tasks_lock = threading.Lock()
        self._workers = []
        self._worker_cond = threading.Condition()
    
    def start_workers(self, num):
        for n in range(0, num):
            self._workers.append(Worker(self))
    
    def stop_workers(self):
        self._stop_workers = True
        
        with self._worker_cond:
            self._worker_cond.notify_all()
        
        for w in self._workers:
            w.join()
        
        self._stop_workers = False
        self._workers = []
        self._tasks = []
    
    def _get_work(self):
        while True:
            if self._stop_workers:
                return None
            
            with self._tasks_lock:
                for task in self._tasks:
                    work = task._get_work()
                    if work is not None:
                        return work
            
            # No work here... let's wait for more.
            with self._worker_cond:
                self._worker_cond.wait()
    
    def add_task(self, task):
        if not task._has_work():
            logging.warning('Added an empty task of type "%s". Ignoring it!', task.__class__.__name__)

            # Make sure it finishes.
            task._done.set()
            task.done.emit()
            return
        
        with self._tasks_lock:
            self._tasks.append(task)
            task._master = self
            task._attached = True

        task.done.connect(self.check_tasks)
        
        with self._worker_cond:
            self._worker_cond.notify_all()

    def check_tasks(self):
        with self._tasks_lock:
            for task in self._tasks[:]:
                if not task._has_work():
                    self._tasks.remove(task)
                    task._attached = False

    def wake_workers(self):
        with self._worker_cond:
            self._worker_cond.notify_all()


class Task(QtCore.QObject):
    _results = None
    _result_lock = None
    _work = None
    _work_count = 0
    _work_lock = None
    _done = None
    _master = None
    _attached = False
    _progress = None
    _progress_lock = None
    _running = 0
    _pending = 0
    _threads = 0
    background = False
    can_abort = True
    aborted = False
    title = None
    done = QtCore.Signal()
    progress = QtCore.Signal(tuple)
    
    def __init__(self, work=None, threads=0):
        super(Task, self).__init__()
        
        if work is None:
            work = []

        self._results = []
        self._work = work
        self._work_count = len(work)
        self._result_lock = threading.Lock()
        self._work_lock = threading.Lock()
        self._done = threading.Event()
        self._progress = dict()
        self._progress_lock = threading.Lock()
        self._threads = threads
    
    def _get_work(self):
        with self._work_lock:
            if len(self._work) == 0:
                return None
            elif self._threads > 0 and self._running >= self._threads:
                return None
            else:
                self._pending += 1
                return (self, (self._work.pop(0),))
    
    def _has_work(self):
        with self._work_lock:
            return len(self._work) > 0
    
    def _init(self):
        with self._progress_lock:
            self._progress[threading.get_ident()] = (0, 'Ready')
            self._running += 1
    
    def _deinit(self):
        with self._progress_lock:
            with self._work_lock:
                self._pending -= 1

            self._progress[threading.get_ident()] = (0, 'Done')
            self._running -= 1
            
            if self._running == 0 and not self._has_work() and not self._done.is_set():
                self._done.set()
                self.done.emit()
    
    def _track_progress(self, prog, text):
        with self._progress_lock:
            self._progress[threading.get_ident()] = (prog, text)
        
        self.progress.emit(self.get_progress())
    
    def post(self, result):
        with self._result_lock:
            self._results.append(result)
    
    def add_work(self, work):
        with self._work_lock:
            self._work.extend(work)
            self._work_count = max(self._work_count, len(self._work))
        
        if self._master is not None:
            if not self._attached:
                self._master.add_task(self)
            else:
                self._master.wake_workers()
    
    def abort(self):
        if not self.can_abort:
            return False
        
        # Empty the work queue, this won't stop running workers but it will
        # stop calls to the work() method.
        with self._work_lock:
            self._work = []
            self.aborted = True

        self._master.check_tasks()
    
    def get_progress(self):
        with self._progress_lock:
            prog = self._progress.copy()
        
        with self._work_lock:
            wc_left = len(self._work)
            wc_total = self._work_count
            pending = self._pending
        
        count = float(wc_total)
        if count == 0:
            count = 0.00001
            total = 1
        else:
            total = (wc_total - wc_left - pending) / count
        
        for item in prog.values():
            total += item[0] * (1.0 / count)

        return total, prog, self.title
    
    def is_done(self):
        if not self._done.is_set():
            with self._progress_lock:
                if self._running == 0 and not self._has_work():
                    self._done.set()

        return self._done.is_set()
    
    def get_results(self):
        if not self._done.is_set():
            self._done.wait()
        
        with self._result_lock:
            return self._results


class MultistepTask(Task):
    _steps = None
    _sdone = False
    _cur_step = -1

    def __init__(self, steps=None, **kwargs):
        super(MultistepTask, self).__init__(**kwargs)

        if steps is None:
            steps = self._steps

        if isinstance(steps, int):
            snum = steps
            steps = []
            for i in range(1, snum + 1):
                steps.append((getattr(self, 'init' + str(i)), getattr(self, 'work' + str(i))))

        self._steps = steps

    def _has_work(self):
        return not self._sdone and not self.aborted

    def _get_work(self):
        with self._work_lock:
            if (self._threads > 0 and self._running >= self._threads) or self._sdone or self.aborted:
                return None
            elif len(self._work) == 0 or self._cur_step < 0:
                if self._pending == 0 and self._running == 0:
                    self._pending += 1
                    return (self, ('MAGIC_MULTITASK_STEP_KEY_###',))
                else:
                    return None
            else:
                self._pending += 1
                return (self, (self._work.pop(0),))

    def work(self, arg):
        # Any better ideas for this magic key?
        if arg == 'MAGIC_MULTITASK_STEP_KEY_###':
            # Maybe we need to advance to the next step.
            self._work_lock.acquire()

            if self._pending == 1 and self._running == 1 and len(self._work) == 0:
                self._work_lock.release()
                self._next_step()
            else:
                # TODO: This still happens on Windows and Mac OS. For some reason it doesn't happen on Linux...
                logging.warning('Either we still have some work to do (unlikely) or there are still some other threads running (%d).', self._running)
                self._work_lock.release()
            
            return

        # Call the current work method
        self._steps[self._cur_step][1](arg)

    def _next_step(self):
        self._cur_step += 1
        logging.debug('Entering step %d of %d in task %s.', self._cur_step + 1, len(self._steps), self.__class__.__name__)

        if self._cur_step >= len(self._steps):
            # That was the last one.
            self._sdone = True
            return

        # Call the init routine.
        self._done.set()
        self._steps[self._cur_step][0]()
        self._done.clear()

        with self._result_lock:
            self._results = []

        # Wake all free workers.
        self._master.wake_workers()


# Curses display
class Textbox(object):
    win = None
    lock = None
    border = False
    content = []
    
    def __init__(self, win, lock=None):
        self.win = win
        self.lock = lock
    
    def wrap(self, text):
        wrapped = []
        y, x, height, width = self.get_coords()
        
        for line in text.split('\n'):
            while len(line) > width:
                wrapped.append(line[:width])
                line = ' ' + line[width:]
            
            wrapped.append(line)
        
        return wrapped
    
    def get_coords(self):
        height, width = self.win.getmaxyx()
        
        if self.border:
            y = x = 1
            height -= 2
            width -= 2
        else:
            y = x = 0
        
        return y, x, height, width
    
    def appendln(self, text):
        with self.lock:
            text = self.wrap(text)
            y, x, height, width = self.get_coords()
            
            self.content.extend(text)
            while len(self.content) > height:
                self.content.pop(0)
            
            self.win.move(y, x)
            self.win.insdelln(-len(text))
            
            start = y + height - len(text)
            self.win.move(start, x)
            for ly in range(start, start + len(text)):
                self.win.addstr(ly, x, text.pop(0))
            
            self.win.refresh()
    
    def append(self, text):
        with self.lock:
            if len(self.content) == 0:
                self.appendln(text)
                return
            
            text = self.wrap(self.content[-1] + text)
            y, x, height, width = self.get_coords()
            self.content[-1] = text.pop(0)

            self.win.addstr(y + height - 1, x, self.content[-1])
            
            if len(text) > 0:
                self.appendln('\n'.join(text))
    
    def set_text(self, text):
        with self.lock:
            text = self.wrap(text)
            self.content = text
            
            if self.border:
                self.win.resize(len(text) + 2, self.win.getmaxyx()[1])
            else:
                self.win.resize(len(text), self.win.getmaxyx()[1])
            
            y, x, height, width = self.get_coords()
            
            self.win.erase()
            if self.border:
                self.win.border()
            
            for ly, line in enumerate(text):
                self.win.addstr(y + ly, x, line)
            
            self.win.refresh()


class CursesOutput(object):
    win = None
    log = None
    other_win = None
    lock = None
    
    def __init__(self, win, log=None, other_win=None):
        self.win = win
        self.log = log
        self.other_win = other_win
        self.lock = threading.RLock()
    
    def write(self, data):
        with self.lock:
            self.win.append(data)
            self.other_win.redrawwin()
            
            if self.log is not None:
                self.log.write(data)
    
    def flush(self):
        with self.lock:
            if self.log is not None:
                self.log.flush()


def _init_curses(scr, cb, log):
    # Setup the display
    height, width = scr.getmaxyx()
    clock = threading.RLock()
    statusw = Textbox(curses.newwin(0, 0, 0, 0), clock)
    statusw.border = True
    win = Textbox(scr, clock)
    
    def show_status(prog, text):
        h, w = statusw.win.getmaxyx()
        statusw.set_text('\n [' + '=' * min(w, int(prog * (w - 8))) + '>\n' + text)
    
    set_callback(show_status)
    
    stdout = sys.stdout
    stderr = sys.stderr
    sys.stderr = sys.stdout = CursesOutput(win, log, statusw.win)
    handlers = []
    
    # Redirect the logging output.
    for handler in logging.getLogger().handlers:
        if isinstance(handler, logging.StreamHandler):
            if handler.stream in (stdout, stderr):
                handlers.append((handler, handler.stream))
                handler.stream = sys.stdout
    
    cb()
    
    for handler, stream in handlers:
        handler.stream = stream
    
    sys.stdout = stdout
    sys.stderr = stderr
    set_callback(None)


def init_curses(cb, log=None):
    curses.wrapper(_init_curses, cb, log)

    # Restore ONLCR mode. (See man tcsetattr for details)
    try:
        import termios
        attr = termios.tcgetattr(sys.stdout)
        attr[1] = attr[1] | termios.ONLCR
        termios.tcsetattr(sys.stdout, termios.TCSAFLUSH, attr)
    except:
        pass



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import logging
import re
import semantic_version

from .qt import QtCore, QtNetwork, QtWebKit
from . import center, api, repo, windows, tasks, util


class WebBridge(QtCore.QObject):

    # NOTE: Update https://github.com/ngld/knossos/wiki/JS%20API whenever you make an API change.
    # getVersion(): str
    #   Returns Knossos's version
    # isFsoInstalled(): bool
    #   Returns true if fs2_open is installed and configured.
    # isFs2PathSet(): bool
    #   Returns true if the path to the FS2 directory is set.
    # selectFs2path(): void
    #   Prompts the user to select their FS2 directory.
    # runGogInstaller(): void
    #   Launches the GOGExtract wizard.
    # getMods(): list of mod objects
    #   Returns a list of all available mods (see the generated section of schema.txt).
    # getInstalledMods(): list of mod objects
    #   Returns a list of all installed mods (see the generated section of schema.txt).
    # getUpdates(): object
    #   Returns a map which follows this syntax: result[mid][local_version] = most_recent_version
    # isInstalled(mid, spec?): bool
    #   Returns true if the given mod is installed. The parameters are the same as query()'s.
    # query(mid, spec?): mod object
    #   Allows the web page to query the local mod cache. The first parameter is the mod ID.
    #   The second parameter is optional and can specify a version requirement. (i.e. ">=3.0.*")
    #   If no matching mod is found, null is returned.
    # fetchModlist(): void
    #   Update the local mod cache.
    # addRepo(name, link): void
    #   Adds a new repository.
    # getRepos(): list of [<name>, <link>]
    #   Returns a list of all configured repositories.
    # install(mid, spec?, pkgs?): int
    #   Installs the given mod. The first two parameters are the same as query()'s.
    #   The third parameter is optional and contains the names of all packages which should be installed (defaults to all required).
    #   Returns -1 if the mod wasn't found, -2 if some of the given packages are missing, 0 if the install failed and 1 on success.
    # uninstall(mid, pkgs?): bool
    #   Uninstalls the given mod. The first parameter is the mod's ID. The second parameter should be used if only some packages should be uninstalled.
    #   Returns true on success.
    # abortDownload(mid): void
    #   Aborts the download for the given mod.
    # runMod(mid): void
    #   Launch fs2_open with the given mod selected.
    # showSettings(mid?): void
    #   Open the settings window for the given mod or fs2_open if no mid is given.
    #   Returns -1 if the mod wasn't found, -2 if the spec is invalid and 1 on success.
    # showPackageList(mid, spec?): int
    #   Open a window which allows the user to install and uninstall packages for the given mod.
    #   Returns -1 if the mod wasn't found, -2 if the spec is invalid and 1 on success.
    # vercmp(a, b): int
    #   Compares two versions

    updateModlist = QtCore.Signal('QVariantMap', str)
    modProgress = QtCore.Signal(str, float, str)

    taskStarted = QtCore.Signal(float, str)
    taskProgress = QtCore.Signal(float, float, str, str)
    taskFinished = QtCore.Signal(float)

    def __init__(self):
        super(WebBridge, self).__init__()

    @QtCore.Slot(result=str)
    def getVersion(self):
        return center.VERSION

    @QtCore.Slot(result=bool)
    def isFsoInstalled(self):
        return api.is_fso_installed()

    @QtCore.Slot(result=bool)
    def isFs2PathSet(self):
        return center.settings['fs2_path'] is not None

    @QtCore.Slot()
    def selectFs2path(self):
        api.select_fs2_path()

    @QtCore.Slot()
    def runGogInstaller(self):
        windows.GogExtractWindow()

    @QtCore.Slot(result='QVariantList')
    def getMods(self):
        return center.mods.get()

    @QtCore.Slot(result='QVariantList')
    def getInstalledMods(self):
        return center.installed.get()

    @QtCore.Slot(result='QVariantMap')
    def getUpdates(self):
        updates = center.installed.get_updates()
        result = {}
        for mid, items in updates.items():
            versions = result[mid] = {}
            for ver_a, ver_b in items.items():
                versions[str(ver_a)] = str(ver_b)

        return result

    @QtCore.Slot(str, result=bool)
    @QtCore.Slot(str, str, result=bool)
    def isInstalled(self, mid, spec=None):
        if spec is None:
            return mid in center.installed.mods
        else:
            spec = util.Spec(spec)
            mod = center.installed.mods.get(mid, None)
            if mod is None:
                return False

            return spec.match(mod.version)

    @QtCore.Slot(str, result='QVariantMap')
    @QtCore.Slot(str, str, result='QVariantMap')
    def query(self, mid, spec=None):
        if spec is not None:
            if spec == '':
                spec = None
            else:
                if re.search(r'^\d+', spec):
                    spec = '==' + spec

                try:
                    spec = util.Spec(spec)
                except:
                    logging.exception('Invalid spec "%s" passed to query()!', spec)
                    return -2

        try:
            return center.mods.query(mid, spec).get()
        except:
            return None

    @QtCore.Slot()
    def fetchModlist(self):
        api.fetch_list()

    @QtCore.Slot(result='QVariantList')
    @QtCore.Slot(bool, result='QVariantList')
    def requestModlist(self, async=False):
        if async:
            center.main_win.update_mod_list()
            return [None]
        else:
            return center.main_win.search_mods()

    @QtCore.Slot(str, str)
    def addRepo(self, repo_url, repo_name):
        repos = center.settings['repos']
        repos.append((repo_url, repo_name))

        api.save_settings()
        if center.main_win is not None:
            center.main_win.update_repo_list()

    @QtCore.Slot(result='QVariantList')
    def getRepos(self):
        return center.settings['repos']

    def _get_mod(self, mid, spec=None, mod_repo=None):
        if spec is not None:
            if spec == '':
                spec = None
            else:
                if re.search(r'^\d+', spec):
                    spec = '==' + spec

                try:
                    spec = util.Spec(spec)
                except:
                    logging.exception('Invalid spec "%s" passed to a web API function!', spec)
                    return -2

        if mod_repo is None:
            mod_repo = center.installed

        try:
            return mod_repo.query(mid, spec)
        except repo.ModNotFound:
            logging.exception('Couldn\'t find mod "%s" (%s)!', mid, spec)
            return -1

    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    def showInfo(self, mid, spec=None):
        mod = self._get_mod(mid, spec, center.mods)
        if mod in (-1, -2):
            return mod

        windows.ModInfoWindow(mod)
        return 0

    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    @QtCore.Slot(str, str, 'QStringList', result=int)
    def install(self, mid, spec=None, pkgs=None):
        mod = self._get_mod(mid, spec, center.mods)
        if mod in (-1, -2):
            return mod

        if pkgs is None:
            pkgs = []
        windows.ModInstallWindow(mod, pkgs)
        return 0

    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    @QtCore.Slot(str, str, 'QStringList', result=int)
    def uninstall(self, mid, spec=None, pkgs=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        if pkgs is None:
            plist = mod.packages
        else:
            plist = []
            pfound = set()
            for pkg in mod.packages:
                if pkg.name in pkgs:
                    plist.append(pkg)
                    pfound.add(pkg.name)

            if len(pfound) < len(pkgs):
                # Some packages are missing
                pmissing = set(pkgs) - pfound
                logging.warning('Missing packages %s.', ', '.join(pmissing))
                return -2

        return api.uninstall_pkgs(plist, name=mod.title)

    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    def updateMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)

        if mod in (-1, -2):
            return mod

        all_vers = list(center.installed.query_all(mid))
        if len(all_vers) == 1:
            # Only one version is installed, let's update it.
            tasks.run_task(tasks.UpdateTask(mod))
        else:
            # Just install the new version
            cur_pkgs = list(mod.packages)
            for i, pkg in enumerate(cur_pkgs):
                cur_pkgs[i] = center.mods.query(mod.mid, None, pkg.name)

            tasks.run_task(tasks.InstallTask(cur_pkgs, cur_pkgs[0].get_mod()))

        center.main_win.update_mod_buttons('progress')
        return 0

    @QtCore.Slot(float)
    def abortTask(self, tid):
        if hasattr(center.main_win, 'abort_task'):
            center.main_win.abort_task(int(tid))

    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    def runMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        api.run_mod(mod)
        return 0

    @QtCore.Slot(result=int)
    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    def showSettings(self, mid=None, spec=None):
        if mid is None:
            windows.SettingsWindow(None)
            return 1
        else:
            mod = self._get_mod(mid, spec)
            if mod in (-1, -2):
                return mod

            windows.ModSettingsWindow(mod)
            return 1

    @QtCore.Slot(str, result=int)
    @QtCore.Slot(str, str, result=int)
    def showPackageList(self, mid=None, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        win = windows.ModSettingsWindow(mod)
        win.show_pkg_tab()

        return 1

    @QtCore.Slot(str, str, result=int)
    def vercmp(self, a, b):
        try:
            a = semantic_version.Version(a)
            b = semantic_version.Version(b)
        except:
            # logging.exception('Someone passed an invalid version to vercmp()!')
            return 0

        return a.__cmp__(b)


class NetworkAccessManager(QtNetwork.QNetworkAccessManager):

    def __init__(self, old_manager):
        super(NetworkAccessManager, self).__init__()

        self.setCache(old_manager.cache())
        self.setCookieJar(old_manager.cookieJar())
        self.setProxy(old_manager.proxy())
        self.setProxyFactory(old_manager.proxyFactory())

    def createRequest(self, operation, request, data):
        if request.url().scheme() == 'fsrs' and operation == self.GetOperation:
            # Rewrite fsrs:// links.
            url = request.url()
            url.setScheme('file')
            path = url.path()

            if path.startswith('/logo'):
                path = path.split('/')
                try:
                    mod = center.mods.query(path[2], util.Spec('==' + path[3]))
                except:
                    try:
                        mod = center.installed.query(path[2], util.Spec('==' + path[3]))
                    except:
                        mod = None

                if mod is not None:
                    logo_path = mod.logo_path
                    if not logo_path:
                        logo_path = mod.logo

                    logging.debug('Rewriting fsrs://%s to %s.', url.path(), logo_path)
                    url.setPath(logo_path)
                else:
                    logging.debug('Mod for path fsrs://%s could not be found!', url.path())
            else:
                url.setPath(os.path.join(center.settings_path, os.path.basename(url.path())))

            request.setUrl(url)

        return super(NetworkAccessManager, self).createRequest(operation, request, data)


class BrowserCtrl(object):
    _view = None
    _bridge = None
    _nam = None

    def __init__(self, webView):
        self._view = webView
        self._bridge = WebBridge()

        settings = webView.settings()
        # settings.setAttribute(QtWebKit.QWebSettings.PluginsEnabled, True)  # May be necessary for YouTube videos...
        settings.setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)
        settings.setAttribute(QtWebKit.QWebSettings.OfflineWebApplicationCacheEnabled, True)

        page = webView.page()
        self._nam = NetworkAccessManager(page.networkAccessManager())
        page.setNetworkAccessManager(self._nam)

        frame = page.mainFrame()
        frame.javaScriptWindowObjectCleared.connect(self.insert_bridge)

        webView.load(QtCore.QUrl('qrc:///html/welcome.html'))

    def insert_bridge(self):
        frame = self._view.page().mainFrame()

        # TODO: Do we have to create a new object here or is there a way around it?
        # The biggest problems are the Signals. I would have to disconnect all their listeners.
        del self._bridge
        self._bridge = WebBridge()
        frame.addToJavaScriptWindowObject('fs2mod', self._bridge)

    def get_bridge(self):
        return self._bridge



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/mod_settings.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(457, 552)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.modTitle = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.modTitle.setFont(font)
        self.modTitle.setAlignment(QtCore.Qt.AlignCenter)
        self.modTitle.setObjectName("modTitle")
        self.verticalLayout.addWidget(self.modTitle)
        self.tabWidget = QtWidgets.QTabWidget(Dialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.modLogo = QtWidgets.QLabel(self.tab)
        self.modLogo.setAlignment(QtCore.Qt.AlignCenter)
        self.modLogo.setObjectName("modLogo")
        self.verticalLayout_2.addWidget(self.modLogo)
        self.modDesc = QtWidgets.QPlainTextEdit(self.tab)
        self.modDesc.setReadOnly(True)
        self.modDesc.setTextInteractionFlags(QtCore.Qt.NoTextInteraction)
        self.modDesc.setObjectName("modDesc")
        self.verticalLayout_2.addWidget(self.modDesc)
        self.tabWidget.addTab(self.tab, "")
        self.flagsTab = QtWidgets.QWidget()
        self.flagsTab.setObjectName("flagsTab")
        self.tabWidget.addTab(self.flagsTab, "")
        self.pkgTab = QtWidgets.QWidget()
        self.pkgTab.setObjectName("pkgTab")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.pkgTab)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.pkgsLayout = QtWidgets.QVBoxLayout()
        self.pkgsLayout.setObjectName("pkgsLayout")
        self.verticalLayout_3.addLayout(self.pkgsLayout)
        self.dlSizeLabel = QtWidgets.QLabel(self.pkgTab)
        self.dlSizeLabel.setObjectName("dlSizeLabel")
        self.verticalLayout_3.addWidget(self.dlSizeLabel)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)
        self.applyPkgChanges = QtWidgets.QPushButton(self.pkgTab)
        self.applyPkgChanges.setObjectName("applyPkgChanges")
        self.verticalLayout_3.addWidget(self.applyPkgChanges)
        self.tabWidget.addTab(self.pkgTab, "")
        self.versionsTab = QtWidgets.QWidget()
        self.versionsTab.setObjectName("versionsTab")
        self.tabWidget.addTab(self.versionsTab, "")
        self.troubleTab = QtWidgets.QWidget()
        self.troubleTab.setObjectName("troubleTab")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.troubleTab)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.checkFiles = QtWidgets.QPushButton(self.troubleTab)
        self.checkFiles.setObjectName("checkFiles")
        self.horizontalLayout.addWidget(self.checkFiles)
        self.delLoose = QtWidgets.QPushButton(self.troubleTab)
        self.delLoose.setObjectName("delLoose")
        self.horizontalLayout.addWidget(self.delLoose)
        self.replaceFiles = QtWidgets.QPushButton(self.troubleTab)
        self.replaceFiles.setObjectName("replaceFiles")
        self.horizontalLayout.addWidget(self.replaceFiles)
        self.verticalLayout_4.addLayout(self.horizontalLayout)
        self.logDisplay = QtWidgets.QTextBrowser(self.troubleTab)
        self.logDisplay.setObjectName("logDisplay")
        self.verticalLayout_4.addWidget(self.logDisplay)
        self.tabWidget.addTab(self.troubleTab, "")
        self.verticalLayout.addWidget(self.tabWidget)

        self.retranslateUi(Dialog)
        self.tabWidget.setCurrentIndex(4)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Mod settings"))
        self.modTitle.setText(_translate("Dialog", "TextLabel"))
        self.modLogo.setText(_translate("Dialog", "TextLabel"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("Dialog", "Infos"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.flagsTab), _translate("Dialog", "Flags"))
        self.dlSizeLabel.setText(_translate("Dialog", "Download size: {DL_SIZE}"))
        self.applyPkgChanges.setText(_translate("Dialog", "Apply changes"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.pkgTab), _translate("Dialog", "Packages"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.versionsTab), _translate("Dialog", "Versions"))
        self.checkFiles.setText(_translate("Dialog", "Check Files"))
        self.delLoose.setText(_translate("Dialog", "Delete loose files"))
        self.replaceFiles.setText(_translate("Dialog", "Replace corrupted files"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.troubleTab), _translate("Dialog", "Troubleshoot"))

